returns <- function(x, ...)
    UseMethod("returns")

## -----[Handling of 'timestamp' and 'period' in methods]-----
##
## Methods are responsible for 'stripping down' the input to x and t,
## calling 'returns.default' (or some other method) and then to
## re-assemble the original class's structure. When there is no
## period, methods should keep timestamp information for themselves,
## not pass it on.

returns.NAVseries <- function(x, period = NULL, complete.first = TRUE,
                              pad = NULL, position = NULL, lag = 1, ...) {

    ## does *not* returns a NAVseries since it is not defined for
    ## returns, only for NAVs (levels)

    if (!is.null(period)) {
        returns.default(x, t = t, period = period,
                        complete.first = complete.first,
                        pad = pad, position = position, lag = lag, ...)
    } else {
        returns.default(x, period = NULL,
                        complete.first = complete.first,
                        pad = pad, position = position, lag = lag, ...)
    }
}

returns.zoo <- function(x, period = NULL, complete.first = TRUE,
                        pad = NULL, position = NULL, lag = 1, ...) {

    t <- time(x)
    x <- coredata(x)

    if (!is.null(period)) {
        returns.default(x, t = t, period = period,
                        complete.first = complete.first,
                        pad = pad, position = position, lag = lag, ...)
    } else {
        ans <- returns.default(x, period = NULL,
                       complete.first = complete.first,
                       pad = pad, position = position, lag = lag, ...)
        attrs <- attributes(ans)
        ans <- if (!is.null(pad))
                   zoo(ans, t)
               else
                   zoo(ans, t[-1L])
        attr(ans, "holdings") <- attrs$holdings
        attr(ans, "contributions") <- attrs$contributions
        ans    
    }
}

returns.data.frame <- function(x, t = NULL, period = NULL,
                               complete.first = TRUE,
                               pad = NULL, position = NULL,
                               weights = NULL, rebalance.when = NULL,
                               lag = 1, ...) {

    ans <- returns.default(x, t = t, period = period,
                           complete.first = complete.first,
                           pad = pad, position = position,
                           weights = weights,
                           rebalance.when = rebalance.when,
                           lag = lag, ...)
    as.data.frame(ans)
}

returns.default <- function(x, t = NULL, period = NULL,
                            complete.first = TRUE,
                            pad = NULL, position = NULL,
                            weights = NULL,
                            rebalance.when = NULL,
                            lag = 1, ...) {

    if (is.null(t) && is.character(period)) {
        warning("no timestamp information available, so ",
                sQuote("period"), " is ignored")
        period <- NULL
    }
    if (!is.null(t) && is.null(period)) {
        warning("timestamp information ignored because ",
                sQuote("period"), " is not specified")
        t <- NULL
    }

    if (is.null(t) &&  is.null(position) && is.null(weights)) {
        .returns(x, pad = pad, lag = lag)
    } else if (is.null(t) &&  is.null(position) && !is.null(weights)) {
        returns_rebalance(prices = x, weights = weights,
                          when = rebalance.when, pad = pad)
    } else if (!is.null(t)) {
        if (is.unsorted(t)) {
            idx <- order(t)
            t <- t[idx]
            x <- x[idx]
        }
        if (!is.null(dim(x)) && min(dim(x)) > 1L)
        if (lag != 1L)
            warning(sQuote("lag"), " is ignored")
        pReturns(x, t, period, complete.first, pad = pad)
    } else {
        if (lag != 1L)
            warning(sQuote("lag"), " is ignored")
        twReturns(x, position, pad = pad)
    }
}

.returns <- function(x, pad = NULL, lag) {
    n <- NROW(x)
    if (n < 2L)
        stop("fewer than two observations")
    do.pad <- !is.null(pad)
    a <- 1:lag
    b <- n:(n-lag+1L)
    if (is.null(dim(x))) {
        x <- as.vector(x)
        rets <- x[-a] / x[-b] - 1
        if (do.pad)
            rets <- c(rep.int(pad, lag), rets)
    } else {
        rets <- x[-a, ,drop = FALSE] / x[-b, ,drop = FALSE] - 1
        if (do.pad)
            rets <- do.call("rbind",
                            c(as.list(rep(pad, lag)), list(rets)))
    }
    rets
}

## not exported
twReturns <- function(price, position, pad = NULL) {
    do.pad <- !is.null(pad)
    position <- as.matrix(position)
    price <- as.matrix(price)
    n <- dim(price)[1L]
    ap <- position*price
    rt <- returns(price)

    M <- price * position
    rsM <- rowSums(M)
    weights <- (M/rsM)
    weights[abs(rsM) < 1e-14] <- 0
    weights <- weights[-n, , drop = FALSE]
    rt <- rowSums(rt * weights)
    if (do.pad)
        rt <- c(pad, rt)
    rt
}

## not exported
pReturns <- function(x, t, period, complete.first = TRUE, pad = NULL) {
    x <- as.matrix(x)
    nc <- ncol(x)
    if (grepl("^ann", period, ignore.case = TRUE)) {
        if (!is.null(pad))
            warning(sQuote("pad"), " is ignored")
        force <- grepl("!$", period)
        ans <- numeric(nc)
        is.ann  <- logical(nc)
        from.to <- array(NA, dim = c(nc, 2))
        colnames(from.to) <- c("from", "to")
        t <- as.Date(t)
        for (j in 1:nc) {
            xj <- x[ ,j]
            t1 <- max(which(!is.na(xj)))
            t0 <- min(which(!is.na(xj)))
            tt <- as.numeric( t[t1] - t[t0] )/365
            tmp <- xj[t1]/xj[t0]
            if (tt > 1 || force) {
                tmp <- tmp^(1/tt)
                is.ann[j] <- TRUE
            }
            ans[j] <- tmp - 1
            from.to[ j, ] <- c(t[t0], t[t1])
        }
        attr(ans, "period") <- if (force)
                                   "annualised!" else "annualised"
        class(from.to) <- "Date"
        attr(ans, "t") <- from.to
        attr(ans, "is.annualised") <- is.ann
    } else if (tolower(period) == "itd") {
        if (!is.null(pad))
            warning(sQuote("pad"), " is ignored")
        ans <- numeric(nc)
        from.to <- array(NA, dim = c(nc, 2))
        colnames(from.to) <- c("from", "to")
        for (j in 1:nc) {
            xj <- x[ ,j]
            t0 <- min(which(!is.na(xj)))
            t1 <- max(which(!is.na(xj)))
            ans[j] <- drop(returns( xj[c(t0, t1)] ))
            from.to[j,] <- c(t[t0], t[t1])
        }
        attr(ans, "period") <- "itd"
        class(from.to) <- "Date"
        attr(ans, "t") <- from.to
    } else if (tolower(period) == "ytd") {
        ## TODO allow syntax like "ytd!" or "ytd02-15"?
        ## (the latter returns a vector of returns ytd
        ## up to 15 Feb).
        ##
        ## ! gives error if when YTD does not make
        ## ! Sys.Date's year
        if (!is.null(pad))
            warning(sQuote("pad"), " is ignored")
        years <- as.numeric(format(t, "%Y"))
        ans <- numeric(nc)
        from.to <- array(NA, dim = c(nc, 2))
        colnames(from.to) <- c("from", "to")
        for (j in 1:nc) {
            xj <- x[ ,j]
            i <- which(years < max(years))
            if (!length(i)) {
                ## all obs are with one year
                i <- seq_along(xj)
                t0 <- min(which(!is.na(xj[i])))
            } else {
                t0 <- max(which(!is.na(xj[i])))
            }
            t1 <- max(which(!is.na(xj)))
            ans[j] <- drop(returns( xj[c(t0, t1)] ))
            from.to[j,] <- c(t[t0], t[t1])
        }
        class(from.to) <- "Date"
        attr(ans, "t") <- from.to
        attr(ans, "period") <- "ytd"
    } else if (tolower(period) == "mtd") {
        if (!is.null(pad))
            warning(sQuote("pad"), " is ignored")
        ymon <- as.numeric(format(t, "%Y%m"))
        ans <- numeric(nc)
        from.to <- array(NA, dim = c(nc, 2))
        colnames(from.to) <- c("from", "to")
        for (j in 1:nc) {
            xj <- x[ ,j]
            i <- which(ymon < max(ymon))
            if (!length(i)) {
                ## all obs are with one month
                i <- seq_along(xj)
                t0 <- min(which(!is.na(xj[i])))
            } else {
                t0 <- max(which(!is.na(xj[i])))
            }
            t1 <- max(which(!is.na(xj)))
            ans[j] <- drop(returns( xj[c(t0, t1)] ))
            from.to[j,] <- c(t[t0], t[t1])
        }
        class(from.to) <- "Date"
        attr(ans, "t") <- from.to
        attr(ans, "period") <- "mtd"
    } else {
        if (length(period) > 1L) {
            by <- period
        } else if (grepl("da(y|i)", period, ignore.case = TRUE)) {
            by <- format(t, "%Y%m%d")
            period <- "daily"
        } else if (grepl("year(ly)?", period, ignore.case = TRUE)) {
            by <- format(t, "%Y")
            period <- "yearly"
        } else if (grepl("month(ly)?", period, ignore.case = TRUE)) {
            by <- format(t, "%Y%m")
            period <- "monthly"
        } else if (grepl("quarter(ly)?", period, ignore.case = TRUE)) {
            by <- paste0(format(t, "%Y"), "-", quarters(t))
            period <- "quarterly"
        } else {
            stop("unknown ", sQuote("period"))
        }
        ii <- last(x, by, TRUE)
        if (complete.first && by[1L] == by[2L])
            ii <- c(1, ii)

        ans <- returns(x[ii, ], pad = pad)
        attr(ans, "t") <- if (is.null(pad)) t[ii][-1L] else t[ii]
        attr(ans, "period") <- period
    }
    class(ans) <- "p_returns"
    ans
}

as.zoo.p_returns <- function (x, ...) {
    if (is.list(x))
        warning(paste0("format of ", sQuote("p_returns"),
                       "objects has changed: see ChangeLog 2015-06-26"))
    zoo(x, attr(x, "t"))
}

## not exported
fmt <- function(x, plus, digits) {
    ans <- format(round(100*x, digits),
                  nsmall = if (digits < 0) 0 else digits,
                  trim = TRUE)
    if (plus)
        paste0(ifelse(x >= 0, "+", ""), ans)
    else
        ans
}

.mtab <- function(x, t, ytd = "YTD", month.names = NULL,
                  zero.print = "0", plus = FALSE, digits = 1,
                  na.print = NULL) {
    years <- as.numeric(format(t, "%Y"))
    mons  <- as.numeric(format(t, "%m"))
    tb <- array("", dim = c(length(unique(years)), 13L))
    tb[cbind(years - years[1L] + 1L, mons)] <- fmt(x, plus, digits)
    for (y in sort(unique(years)))
        tb[y - years[1L] + 1L, 13L] <- fmt(prod(x[years==y & !is.na(x)] + 1L) - 1L,
                                           plus, digits)
    rownames(tb) <- sort(unique(years))
    colnames(tb) <- if (is.null(month.names))
                        c(format(as.Date(paste0("2012-", 1:12, "-1")), "%b"), ytd)
                    else
                        c(month.names, ytd)
    if (zero.print != "0")
        tb <- gsub("^ *[+-]?[0.]+$", zero.print, tb)
    tb
}

## not exported
print.p_returns <- function(x, ..., year.rows = TRUE,
                           month.names = NULL, zero.print = "0", plus = FALSE,
                           digits = 1, na.print = NULL) {
    if (is.list(x))
        warning("format of 'p_returns' objects has changed:",
                " see ChangeLog 2015-06-26")
    period <- attr(x, "period")
    timestamp <- attr(x, "t")
    if (period == "monthly" && is.null(dim(x))) {
        if (year.rows)
            print(.mtab(x, timestamp, ytd = "YTD", month.names = month.names,
                        zero.print = zero.print, plus = plus, digits = digits),
                  quote = FALSE, print.gap = 1, right = TRUE)
        else
            print(t(.mtab(x, timestamp, ytd = "YTD", month.names = month.names,
                       zero.print = zero.print, plus = plus, digits = digits)),
                  quote = FALSE, print.gap = 2, right = TRUE)
    } else if (period == "monthly") {
        tmp <- x
        tmp <- format(round(tmp*100,1), nsmall = digits)
        row.names(tmp) <- as.character(timestamp)
        print(unclass(tmp), quote = FALSE, print.gap = 2)
    } else if (period == "yearly") {
        tmp <- x
        names(tmp) <- format(timestamp, "%Y")
        if (year.rows)
            print(fmt(tmp, plus, digits), quote = FALSE)
        else
            print(as.matrix(fmt(tmp, plus, digits)), quote = FALSE)
    } else if (grepl("annualised", period)) {
        r_str <- paste0(format(round(x*100, digits), nsmall = digits), "%  ")
        cal_str <- paste0("[",
                       format(timestamp[,1],"%d %b %Y"), " -- ",
                       format(timestamp[,2],"%d %b %Y"))

        note <- rep("]", length(x))
        note[as.numeric(timestamp[,2L]-timestamp[,1L])/365 < 1 &
              attr(x, "is.annualised")] <-
            "; less than one year, but annualised]"
        note[as.numeric(timestamp[,2L]-timestamp[,1L])/365 < 1 &
              !attr(x, "is.annualised")] <-
            "; less than one year, not annualised]"
        cat(paste0(r_str, cal_str, note, collapse = "\n"), "\n")
    } else if (period == "ytd" || period == "mtd" || period == "itd") {
        r_str <- paste0(format(round(x*100, digits), nsmall = digits), "%  ")
        cal_str <- paste0("[",
                          format(timestamp[,1],"%d %b %Y"), " -- ",
                          format(timestamp[,2],"%d %b %Y"),
                          "]")
        cat(paste0(r_str, cal_str, collapse = "\n"), "\n")
    } else {
        print(unclass(x))
    }
    invisible(x)
}

## not exported
toLatex.p_returns <- function(object, ..., year.rows = TRUE,
                              ytd = "YTD", month.names = NULL, eol = "\\\\",
                              stand.alone = FALSE) {
    if (is.list(object))
        warning("format of 'p_returns' objects has changed: see ChangeLog 2015-06-26")

    period <- attr(object, "period")
    timestamp <- attr(object, "t")

    if (grepl("month(ly)?", period, ignore.case = TRUE)) {
        if (year.rows)
            mt <- .mtab(object, timestamp, ytd = ytd, month.names = month.names)
        else
            mt <- t(.mtab(object, timestamp, ytd = ytd, month.names = month.names))
    } else {
        stop("currently only supported for period ", sQuote("month"))
    }
    mt <- rbind(colnames(mt), mt)
    mt <- cbind(rownames(mt), mt)
    mt <- paste(apply(mt, 1, function(x) paste(x, collapse = "&")), eol)
    class(mt) <- "Latex"
    mt
}

## not exported
toHTML.p_returns <- function(x, ..., year.rows = TRUE,
                            ytd = "YTD", month.names = NULL,
                            stand.alone = TRUE,
                            table.style = NULL,
                            table.class = NULL,
                            th.style = NULL,
                            th.class = NULL,
                            td.style = "text-align:right; padding:0.5em;",
                            td.class = NULL,
                            tr.style = NULL, tr.class = NULL,
                            browse = FALSE) {

    if (is.list(x))
        warning("format of 'p_returns' objects has changed: see ChangeLog 2015-06-26")

    period <- attr(x, "period")
    timestamp <- attr(x, "t")

    .ctag <- function(value, tag)
        if (!is.null(value) && value != "")
            paste0(" ", tag, "=", value)
        else
            character(0L)

    .th <- function(x, style = th.style, class = th.class){
        open <- paste0("<th", .ctag(style, "style"), .ctag(class, "class"), ">")
        paste0(open, x, "</th>")
    }
    .td <- function(x, style = td.style, class = td.class){
        open <- paste0("<td", .ctag(style, "style"), .ctag(class, "class"), ">")
        paste0(open, x, "</td>")
    }
    .tr <- function(x, style = tr.style, class = tr.class){
        open <- paste0("<tr", .ctag(style, "style"), .ctag(class, "class"), ">")
        paste0(open, x, "</tr>")
    }

    if (grepl("month(ly)?", period)) {
        if (year.rows)
            mt <- .mtab(x, timestamp, ytd = ytd, month.names = month.names)
        else
            mt <- t(.mtab(x, timestamp, ytd = ytd, month.names = month.names))
    } else {
        stop("currently only supported for period ", sQuote("month"))
    }
    mt <- rbind(colnames(mt), mt)
    mt <- cbind(rownames(mt), mt)
    mt <- unname(mt)
    mt[1, ]   <- .th(mt[1, ])
    mt[-1 ,1] <- .th(mt[-1, 1])
    mt[-1,-1] <- .td(mt[-1,-1])

    mt <- .tr(apply(mt, 1, paste, collapse = ""))
    open <- paste0("<table", .ctag(table.style, "style"),
                             .ctag(table.class, "class"), ">")
    mt <- c(open, mt, "</table>")

    if (browse) {
        tmp <- tempfile("PMwR_", fileext = ".html")
        writeLines(mt, con= file(tmp), sep = "\n")
        browseURL(paste0("file://", tmp))
        invisible(mt)
    } else
        mt
}

## not exported
toText.p_returns <- function(x, ..., year.rows = TRUE,
                            ytd = "YTD", month.names = NULL) {

    if (is.list(x))
        warning("format of 'p_returns' objects has changed: see ChangeLog 2015-06-26")

    period <- attr(x, "period")
    timestamp <- attr(x, "t")

    if (grepl("month(ly)?", period)) {
        if (year.rows)
            mt <- .mtab(x, timestamp, ytd = ytd, month.names = month.names)
        else
            mt <- t(.mtab(x, timestamp, ytd = ytd, month.names = month.names))
    } else {
        stop("currently only supported for period ", sQuote("month"))
    }
    mt <- rbind(colnames(mt), mt)
    mt <- cbind(rownames(mt), mt)
    mt <- unname(mt)
    mt <- apply(mt, 2, function(x) {format(x,
                                           width = max(5, nchar(x)),
                                           justify = "right")})
    mt <- apply(mt, 1, paste0, collapse = "")
    class(mt) <- "text"
    mt
}

## not exported
returns_rebalance <- function(prices, weights, when = NULL, pad = NULL) {
    nr <- nrow(prices)

    if (is.null(dim(prices)))
        stop("prices needs to be a matrix")

    if (nr < 2L)
        stop("less than 2 rows in prices: cannot compute returns")

    if (is.null(dim(weights)) && is.null(when)) {
        ## TODO faster implementation
    }

    if (is.null(dim(weights)))
        weights <- array(1, dim = c(nr, 1)) %*% weights

    if (dim(weights)[2L] != ncol(prices))
        warning("length of weights does not match number of price series")

    if (is.null(when) || isTRUE(when))
        when <- rep(TRUE, nr)

    if (is.numeric(when)) { ## cases such as '1,4,12,23,...'
        tmp <- logical(nr)
        tmp[round(when)] <- TRUE
        when <- tmp
    }
    val <- numeric(nr)
    h <- ctb <- array(0, dim = dim(prices))

    val[1] <- 1
    h[1, ] <- weights[1, ]/prices[1, ]

    for (i in 2:nr) {
        val[i] <- sum(prices[i, ]*h[i - 1, ])
        ctb[i,] <- (prices[i,]-prices[i-1,]) * h[i-1,] / val[i-1]
        if (when[i])
            h[i, ] <- val[i] * weights[i, ]/prices[i, ]
        else
            h[i, ] <- h[i - 1, ]
    }
    ans <- returns(val, pad = pad)
    attr(ans, "holdings") <- h
    attr(ans, "contributions") <- ctb
    ans
}

rc <- function(R, weights, timestamp, segments = NULL) {
    if (missing(weights))
        weights <- 1
    if (is.null(segments)) {
        segments <- if (!is.null(cr <- colnames(R)))
                        cr
                    else if (!is.null(cr <- colnames(weights)))
                        cr
                    else 
                        paste0("segment_", 1:ncol(weights))
    }
    if (missing(timestamp))
        timestamp <- 1:nrow(R)
    ns <- length(segments)
    nt <- length(timestamp)
    df <- data.frame(timestamp,
                     cbind(weights*R, rowSums(weights*R)),
                     stringsAsFactors = FALSE)
    names(df) <- c("timestamp", segments, "total")

    later_r <- c(rev(cumprod(1 + rev(df[["total"]])))[-1], 1)
    
    total <- rep(NA_real_, ns + 1)
    names(total) <- c(segments, "total")
    for (i in seq_len(ns))
        total[[i]] <- sum(df[[i + 1]] * later_r)
    total[[ns + 1]] <- cumprod(df[["total"]] + 1)[[nt]] - 1
    list(period_contributions = df,
         total_contributions = total)
}

if (FALSE) {

    
    prices <- c(100 ,102 ,104 ,104 ,104.5 ,
                2   ,2.2 ,2.4 ,2.3 ,2.5   ,
                3.5 ,3   ,3.1 ,3.2 ,3.1)
    
    dim(prices) <- c(5,3)
    
    weights <- c(1,0,0)
    when <- as.logical(c(1,0,1,0,1))
    when <- as.logical(c(1,1,1,1,1))
    when <- TRUE
    
    
    
    ## TESTS
    weights <- c(1,0,0)
    all.equal(c(returns.rebalance(prices, weights, when = when, aggregate = TRUE, pad = 0)),
              c(returns(prices[,1],pad=0)))
    weights <- c(0,1,0)
    all.equal(c(returns.rebalance(prices, weights, when = when, aggregate = TRUE, pad = 0)),
              c(returns(prices[,2],pad=0)))
    weights <- c(0,0,1)
    all.equal(c(returns.rebalance(prices, weights, when = when, aggregate = TRUE, pad = 0)),
              c(returns(prices[,3],pad=0)))
    weights <- c(.5,.4,.1)
    all.equal(c(returns(prices,pad=0) %*% weights),
              c(returns.rebalance(prices, weights, when = when, aggregate = TRUE, pad = 0)))
    
    when <- as.logical(c(1,0,0,0,0))
    !isTRUE(all.equal(
         c(returns(prices,pad=0) %*% weights),
         c(returns.rebalance(prices, weights, when = when, aggregate = TRUE, pad = 0))))
    
    
    
    ## not exported
    prices <- c(100 ,102 ,104 ,104 ,104.5 ,
                2   ,2.2 ,2.4 ,2.3 ,2.5   ,
                3.5 ,3   ,3.1 ,3.2 ,3.1)
    
    dim(prices) <- c(5,3)
    
    weights <- c(1,0,0)
    when <- as.logical(c(1,1,1,1,1))
    when <- as.logical(c(1,0,1,0,1))
    
    returns_rebalance2 <- function(prices, weights, when = NULL, pad = NULL) {
        prices <- as.matrix(prices)
        
        nr <- nrow(prices)
        pos <- array(NA, dim = dim(prices))
        ## w <- array(NA, dim = dim(prices))
        
        if (is.null(when && identical(when, TRUE)))
            when <- seq_len(nc)
        else if (is.logical(when))
            when <- which(when)
    
    
    for (i in when)
        pos[i, ] <- weights / prices[i, ] ## TODO: if w is matrix, w[i, ]
    ## TODO na.locf
    
    pos <- na.locf(pos)
    pos[is.na(pos)] <- 0
    w <- pos * prices
    w <- w/rowSums(w)
    
    rowSums(returns(prices) * w[-nr,])
    }
    
    returns_rebalance2(prices, weights, when = TRUE)
    returns_rebalance(prices, weights, when = TRUE)
    
    require("rbenchmark")
    benchmark(returns_rebalance2(prices, weights, when = TRUE),
              returns_rebalance(prices, weights, when = TRUE),
              columns = c("test", "relative", "elapsed"))
    

}
